// Copyright (C) 2008, 2009, 2010 EPITA Research and Development
// Laboratory (LRDE)
//
// This file is part of the Milena Library.  This library is free
// software; you can redistribute it and/or modify it under the terms
// of the GNU General Public License version 2 as published by the
// Free Software Foundation.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this library; see the file COPYING.  If not, write to
// the Free Software Foundation, 51 Franklin Street, Fifth Floor,
// Boston, MA 02111-1307, USA.
//
// As a special exception, you may use this file as part of a free
// software library without restriction.  Specifically, if other files
// instantiate templates or use macros or inline functions from this
// file, or you compile this file and link it with other files to
// produce an executable, this file does not by itself cause the
// resulting executable to be covered by the GNU General Public
// License.  This exception does not however invalidate any other
// reasons why the executable file might be covered by the GNU General
// Public License.

/// \file apps/mesh-segm-skel/mesh-segm.cc
/// \brief A program performing a WST-based segmentation of the
/// surface of the (triangle) mesh of a statue.

// FIXME: Factor commons parts between mesh-segm and mesh-skel.

#include <cstdlib>
#include <cmath>

#include <utility>
#include <iostream>

#include <TriMesh.h>

#include <mln/core/alias/point3d.hh>

#include <mln/fun/i2v/array.hh>
#include <mln/convert/from_to.hh>

#include <mln/util/graph.hh>
#include <mln/util/line_graph.hh>
#include <mln/util/site_pair.hh>
#include <mln/core/image/edge_image.hh>
#include <mln/make/edge_image.hh>

#include <mln/morpho/closing/area.hh>
#include <mln/morpho/watershed/flooding.hh>

#include "io.hh"


// Doesn't C++ have a better way to express Pi ?
const float pi = 4 * atanf(1);


int main(int argc, char* argv[])
{
  if (argc != 4)
    {
      std::cerr << "usage: " << argv[0] << " input.off lambda output.off"
		<< std::endl;
      std::exit(1);
    }

  std::string input_filename = argv[1];
  unsigned lambda = atoi(argv[2]);
  std::string output_filename = argv[3];


  /*-------.
  | Mesh.  |
  `-------*/

  // TriMesh is a pain: it systematically allocates on the heap.
  // Introduce another name to manipulate the mesh as a (non-pointer)
  // object.
  TriMesh* mesh_ptr = TriMesh::read(input_filename.c_str());
  if (!mesh_ptr)
    std::exit(2);
  TriMesh& mesh = *mesh_ptr;

  // Computes faces (triangles).
  mesh.need_faces();
  // Computation of the mean curvature on each vertex of the mesh.
  mesh.need_curvatures();
  /* FIXME: Our implementation of the WST doesn't work well with
     floats (yet).  Convert floating point values to a proportional
     integer value for the moment.  */
  typedef int curv_t;
  std::vector<curv_t> vertex_h_inv(mesh.vertices.size(), 0.f);
  for (unsigned v = 0; v < mesh.vertices.size(); ++v)
    {
      float h = (mesh.curv1[v] + mesh.curv2[v]) / 2;
      float h_inv = 1 / pi * atan(-h) + pi / 2;
      /* FIXME: This coefficient is used to distinguish small
         curvature values.  We should get rid of it as soon as
         morpho::meyer_wst works correctly on images of float
         values.  */
      vertex_h_inv[v] = 1000 * h_inv;
    }

  /*--------.
  | Graph.  |
  `--------*/

  /* Build a graph whose vertices correspond to the faces of the mesh,
     whose edges (between two vertices) correspond to edges (between
     two faces) of the mesh.  */

  // Values associated to the graph sites.
  // (Reminder: currently, values associated to vertices in a
  // line_graph_image are dummy.)
  std::vector<curv_t> vertex_values (mesh.faces.size(), 0.f);
  std::vector<curv_t> edge_values;

  mln::util::graph g;
  // Populate the graph with vertices.
  g.add_vertices(mesh.faces.size());

  // Populate the graph with edges.
  mesh.need_across_edge();
  for (unsigned f = 0; f < mesh.faces.size(); ++f)
    for (unsigned e = 0; e < 3; ++e)
      {
	int f_adj = mesh.across_edge[f][e];
	if (f_adj != -1)
	  {
	    // Add an edge into the graph.
	    mln::util::edge_id_t new_edge = g.add_edge(f, f_adj);
	    mln_invariant(new_edge.is_valid());

	    /* Among the edges adjacent to F and F_ADJ, find the one
	       they share.  In practice, find the two vertices shared
	       by F and F_ADJ.  */
	    /* FIXME: We lack a proper interface from the TriMesh
	       structure to do this elegantly.  */
	    std::vector<int> adj_vertices;
	    adj_vertices.reserve(2);
	    for (unsigned i = 0; i < 3; ++i)
	      for (unsigned j = 0; j < 3; ++j)
		if (mesh.faces[f][i] == mesh.faces[f_adj][j])
		  adj_vertices.push_back(mesh.faces[f][i]);
	    mln_assertion(adj_vertices.size() == 2);

	    // Compute the mean curvature on the edge.
	    edge_values.push_back((vertex_h_inv[adj_vertices[0]] + 
				   vertex_h_inv[adj_vertices[1]])
				  / 2);

	    // Check the consistency of the two arrays.
	    mln_assertion(g.edges().size() == edge_values.size());
	  }
      }

  /*-------------------.
  | Line graph image.  |
  `-------------------*/

  // Points associated to the line graph vertices (edges in the graph).
  typedef mln::util::site_pair<mln::point3df> P;
  typedef mln::fun::i2v::array<P> fsite_t;
  fsite_t sites(g.e_nmax());
  for (unsigned i = 0; i < g.e_nmax(); ++i)
    /* FIXME: Add a dummy location information for the moment.  This
       doesn't really matter as we won't use it to produce the output.
       However, we may need it later: do not forget to add relevant
       locations (obtained from the mesh) in a future version of this
       program.  */
    sites(i) = P(mln::point3df(0, 0, 0), mln::point3df(1, 1, 1));

  // FIXME: This intermediate structure costs us a copy.
  typedef mln::fun::i2v::array<curv_t> values_t;
  values_t values;
  mln::convert::from_to(edge_values, values);

  typedef mln::edge_image<P, curv_t, mln::util::graph> ima_t;
  ima_t graph_ima(g, sites, values);


  /*-----------------.
  | Simplification.  |
  `-----------------*/

  typedef ima_t::nbh_t nbh_t;
  nbh_t nbh;

  // ``Area'' is not a proper term here, since it is defined to the
  // number of edges in the graph.  Taking into account the actual
  // area of triangles would be a more clever choice.
  ima_t closed_graph_ima = mln::morpho::closing::area(graph_ima, nbh, lambda);

  /*------.
  | WST.  |
  `------*/

  typedef unsigned wst_val_t;
  wst_val_t nbasins;
  typedef mln_ch_value_(ima_t, wst_val_t) wst_ima_t;
  wst_ima_t wshed =
    mln::morpho::watershed::flooding(closed_graph_ima, nbh, nbasins);
  std::cout << "nbasins = " << nbasins << std::endl;

  /*------------------------------------------.
  | Label graph vertices (i.e., mesh faces).  |
  `------------------------------------------*/

  // FIXME: Use a vertex_image and a e2v neighborhood instead?

  /* FIXME: p_edges does not provide nvertices(), so we cannot write

       std::vector<unsigned> vertex_label(wshed.domain().nvertices(), 0);

     Likewise, p_vertices does not provide n_edges(). */
  std::vector<unsigned> vertex_label(wshed.domain().graph().v_nmax(), 0);
  mln_piter_(wst_ima_t) pw(wshed.domain());
  for_all(pw)
    if (wshed(pw) != 0)
      {
	mln_psite_(wst_ima_t) pp(pw);
	vertex_label[pp.v1().id()] = wshed(pw);
	vertex_label[pp.v2().id()] = wshed(pw);
      }

  /*---------.
  | Output.  |
  `---------*/

  // Choose random colors for each basin number.
  std::vector<mln::value::rgb8> basin_color (nbasins + 1);
  for (unsigned i = 0; i <= nbasins; ++i)
    basin_color[i] = mln::value::rgb8(random() % 256,
				      random() % 256,
				      random() % 256);

  // Assign colors to graph vertices (mesh faces).
  std::vector<mln::value::rgb8> face_color (vertex_label.size());
  for (unsigned i = 0; i < vertex_label.size() ; ++i)
    face_color[i] = basin_color[vertex_label[i]];

  // Taken and adapted from TriMesh_io.cc
  FILE* f_out = fopen(output_filename.c_str(), "wb");
  if (!f_out)
    {
      std::cerr << "Error opening " << output_filename.c_str()
		<< " for writing." << std::endl;
      std::exit(2);
    }
  write_off_colored(mesh_ptr, face_color, f_out);
  fclose(f_out);

  delete mesh_ptr;
}
